package com.grapeshot.halfnes;

import com.grapeshot.halfnes.mappers.BadMapperException;
import com.grapeshot.halfnes.mappers.Mapper;
import java.io.*;
import java.util.prefs.Preferences;

/**
 *
 * @author Andrew Hoffman
 */
public class NES {

    private final Preferences prefs = Preferences.userNodeForPackage(this.getClass());
    private Mapper mapper;
    private APU apu;
    private CPU cpu;
    private CPURAM cpuram;
    private PPU ppu;
    private Controller controller1;
    private Controller controller2;
    final public static String VERSION = "0.036";
    private boolean runEmulation, dontSleep = false;
    private long frameStartTime = 0;
    public long framecount = 0;
    private long frameDoneTime = 0;
    private long sleepingtest = 0;
    private boolean frameLimiterOn = true;
    final public static long FRAME_NS = 16639267;
    private File currentrom;
    private final GUI gui = new GUI(this);

    public NES() {
        java.awt.EventQueue.invokeLater(new Runnable() {

            public void run() {
                gui.run();
            }
        });
    }

    public void run(final File romtoload) {
        Thread.currentThread().setPriority(Thread.NORM_PRIORITY + 1);
        //set thread priority higher than the interface thread
        currentrom = romtoload;
        loadROM(romtoload);
        run();
    }

    public void run() {
        controller1 = new Controller(prefs, 0);
        controller2 = new Controller(prefs, 1);
        gui.addKeyListener(controller1);
        gui.addKeyListener(controller2);
        forceHighResolutionTimer();
        while (true) {
            if (runEmulation) {
                frameStartTime = System.nanoTime();
                runframe();
                if (frameLimiterOn && !dontSleep) {
                    sleep();
                }
                frameDoneTime = System.nanoTime() - frameStartTime;
            } else {
//                if (ppu != null) {
//                    //keep doing screen updates anyway.
//                    gui.setFrame(ppu.renderFrame());
//                }
                //turns out that locks up the event thread.
                //TODO find a better solution
                try {
                    Thread.sleep(16);
                } catch (InterruptedException e) {
                }
            }
        }
    }

    public synchronized void runframe() {
        final int scanlinectrfire = 256;
        //the main method sequencing everything that has to happen in the nes each frame
        //loops unrolled a bit to avoid some conditionals every cycle
        //vblank
        //start by setting nmi
        if ((utils.getbit(ppu.ppuregs[0], 7) && framecount > 1)) {
            //^hack, without this Lolo 2 and/or Vegavox NSFs don't work
            cpu.runcycle(241, 9000);
            cpu.nmi();
            // do the nmi but let cpu run ONE extra instruction first
            // (fixes Adventures of Lolo 2, Milon's Secret Castle, Solomon's Key)
        }
        for (int scanline = 241; scanline < 261; ++scanline) {
            //most of vblank period
            cpu.cycle(scanline, scanlinectrfire);
            mapper.notifyscanline(scanline);
            cpu.cycle(scanline, 341);
        }
        //scanline 261 
        //turn off vblank flag
        ppu.ppuregs[2] &= 0x80;
        cpu.cycle(261, 30);
        // turn off sprite 0, sprite overflow flags
        ppu.ppuregs[2] &= 0x9F;
        cpu.cycle(261, scanlinectrfire);
        mapper.notifyscanline(261);
        cpu.cycle(261, (((framecount & 1) == 1) && utils.getbit(ppu.ppuregs[1], 3)) ? 340 : 341);
        //odd frames are shorter by one PPU pixel if rendering is on.

        dontSleep = apu.bufferHasLessThan(1000);
        //if the audio buffer is completely drained, don't sleep for this frame
        //this is to prevent the emulator from getting stuck sleeping too much
        //on a slow system or when the audio buffer runs dry.

        apu.finishframe();
        cpu.modcycles();
        //active drawing time
        for (int scanline = 0; scanline < 240; ++scanline) {
            if (!ppu.drawLine(scanline)) {
                cpu.cycle(scanline, scanlinectrfire);
                mapper.notifyscanline(scanline);
            }else{
                //it is de sprite zero line
                final int sprite0x = ppu.sprite0x;
                if (sprite0x < scanlinectrfire) {
                    cpu.cycle(scanline, sprite0x);
                    ppu.ppuregs[2] |= 0x40; //sprite 0 hit
                    cpu.cycle(scanline, scanlinectrfire);
                    mapper.notifyscanline(scanline);
                } else {
                    cpu.cycle(scanline, scanlinectrfire);
                    mapper.notifyscanline(scanline);
                    cpu.cycle(scanline, sprite0x);
                    ppu.ppuregs[2] |= 0x40; //sprite 0 hit
                }
            }
            //and finish out the scanline
            cpu.cycle(scanline, 341);
        }
        //scanline 240: dummy fetches
        cpu.cycle(240, scanlinectrfire);
        mapper.notifyscanline(240);
        cpu.cycle(240, 341);
        //set the vblank flag
        ppu.ppuregs[2] |= 0x80;
        //render the frame
        gui.setFrame(ppu.renderFrame());
        if ((framecount & 2047) == 0) {
            //save sram every 30 seconds or so
            saveSRAM(true);
        }
        ++framecount;
    }

    public void sleep() {

        //Frame Limiter
        final long timeleft = System.nanoTime() - frameStartTime;
        if (timeleft < FRAME_NS) {
            final int sleepytime = (int) (FRAME_NS - timeleft + sleepingtest);
            if (sleepytime < 0) {
                return;
                //don't sleep at all.
            }
            sleepingtest = System.nanoTime();
            try {
                Thread.sleep(sleepytime / 1000000, (sleepytime % 1000000) / 1000);
                // sleep for rest of the time until the next frame
            } catch (InterruptedException ex) {
            }
            sleepingtest = System.nanoTime() - sleepingtest;
            //now sleeping test has how many ns the sleep *actually* was
            sleepingtest = sleepytime - sleepingtest;
            //now sleepingtest has how much the next frame needs to be delayed by to make things match
        }
    }

    public void toggleFrameLimiter() {
        if (frameLimiterOn) {
            frameLimiterOn = false;
        } else {
            frameLimiterOn = true;
        }
    }

    public synchronized void loadROM(final File file) {
        if (apu != null) {
            //if rom already running save its sram before closing
            apu.destroy();
            saveSRAM(false);
        }
        runEmulation = false;
        frameLimiterOn = true;
        final ROMLoader loader = new ROMLoader(file);
        try {
            if (utils.getExtension(file).equalsIgnoreCase(".nes")) {
                loader.parseInesheader();
                mapper = Mapper.getCorrectMapper(loader.mappertype);
                mapper.setLoader(loader);
                mapper.loadrom();
                //now some annoying getting of all the references where they belong
                cpuram = mapper.getCPURAM();
                cpu = mapper.cpu;
                ppu = mapper.ppu;
                apu = new APU(this, cpu, cpuram);
                cpuram.setAPU(apu);
                cpuram.setPPU(ppu);
                currentrom = file;

                //} else if (utils.getExtension(f).equalsIgnoreCase(".nsf")) {
                //not done yet, this needs seious refactoring first!
            } else {
                throw new IOException("Unrecognized extension");
            }
        } catch (BadMapperException e) {
            gui.messageBox("Error Loading File: ROM is"
                    + " corrupted or uses an unsupported mapper.\n");
            runEmulation = false;
            return;

        } catch (IOException q) {
            gui.messageBox("Error loading ROM: " + q.getMessage());
            runEmulation = false;
            return;
        }
        framecount = 0;
        //if savestate exists, load it
        if (mapper.hasSRAM()) {
            loadSRAM();
        }
        //and start emulation

        cpu.init();
        runEmulation = true;
    }

    private void saveSRAM(final boolean async) {
        if (mapper != null && mapper.hasSRAM() && mapper.supportsSaves()) {

            if (async) {
                utils.asyncwritetofile(mapper.getPRGRam(), new File(utils.stripExtension(currentrom) + ".sav"));
            } else {
                utils.writetofile(mapper.getPRGRam(), new File(utils.stripExtension(currentrom) + ".sav"));
            }
        }
    }

    private void loadSRAM() {
        final File sram = new File(utils.stripExtension(currentrom) + ".sav");
        if (sram.isFile() && mapper.supportsSaves() && (sram.length() == 8192)) {
            mapper.setPRGRAM(utils.readfromfile(sram));
        }

    }

    public static void forceHighResolutionTimer() {
        //UGLY HACK ALERT: Just realized why sleep() rounds to nearest
        //multiple of 10: it's because no other program is using high resolution timer.
        //This should, hopefully, fix that.
        final Thread daemon = new Thread("ForceHighResolutionTimer") {

            @Override
            public void run() {
                while (true) {
                    try {
                        Thread.sleep(99999999999L);
                    } catch (InterruptedException e) {
                        break;
                    }
                }
            }
        };
        daemon.setDaemon(true);
        daemon.start();
    }

    public void quit() {
        //save SRAM and quit
        if (cpu != null && currentrom != null) {
            runEmulation = false;
            saveSRAM(false);
        }
        System.exit(0);
    }

    public synchronized void reset() {
        if (cpu != null) {
            cpu.reset();
            runEmulation = true;
            apu.stopAudio();
            apu.startAudio();
        }
        //reset frame counter as well because PPU is reset
        //on Famicom, PPU is not reset when Reset is pressed
        //but some NES games expect it to be and you get garbage.
        framecount = 0;
    }
    
        public synchronized void reloadROM() {
            loadROM(currentrom);
    }

    public synchronized void pause() {
        if (apu != null) {
            apu.stopAudio();
        }
        runEmulation = false;
    }

    public long getFrameTime() {
        return frameDoneTime;
    }

    public String getrominfo() {
        if (mapper != null) {
            return mapper.getrominfo();
        }
        return null;
    }

    public Preferences getPrefs() {
        return prefs;
    }

    public synchronized void frameAdvance() {
        runEmulation = false;
        if (cpu != null) {
            runframe();
        }
    }

    public synchronized void resume() {
        if (apu != null) {
            apu.startAudio();
        }
        if (cpu != null) {
            runEmulation = true;
        }
    }

    public String getCurrentRomName() {
        return currentrom.getName();
    }

    public boolean isFrameLimiterOn() {
        return frameLimiterOn;
    }

    public void messageBox(final String string) {
        gui.messageBox(string);
    }

    public Controller getcontroller1() {
        return controller1;
    }

    public Controller getcontroller2() {
        return controller2;
    }
}
